#define AUTO_DLLMAIN
#include <DynRPG/DynRPG.h>
//#include <emmintrin.h>
#include "ImageData.h"

#include <sstream>

typedef std::map<int, int> Active;
typedef std::map<int, ImageData*> Blend;
Blend blends;
Active active;

void drawAdditive(int id)
{
    ImageData* data = blends[id];
    RPG::Picture* pic = data->pic;

    float alpha = 1 - 0.01*pic->transparency;

    int xwid = data->width;
    int ywid = data->height;
    int swid = RPG::screen->canvas->lineSize/2;

    int x = pic->x - xwid/2;
    int y = pic->y - ywid/2;

    int x_t;
    int y_t;

    int x_c;
    int y_c;

    int x_e;
    int y_e;

    if(x < 0)
    {
        x_t = 0;
        x_c = -x;
        int tmp = xwid + x;
        x_e = (tmp > 320)?320:tmp;
    }
    else
    {
        x_t = x;
        x_c = 0;
        x_e = (x + xwid > 320)?320-x:xwid;
    }

    if(y < 0)
    {
        y_t = 0;
        y_c = -y;
        int tmp = ywid + y;
        y_e = (tmp > 240)?240:tmp;
    }
    else
    {
        y_t = y;
        y_c = 0;
        y_e = (y + ywid > 240)?240-y:ywid;
    }

    uint16_t* scanline = RPG::screen->canvas->getScanline(0);

    int sPtr = x_t + y_t * swid;
    int iPtr = x_c + y_c * xwid;

    for(int i = 0; i < y_e; ++i)
    {
        for(int j = 0; j < x_e; ++j)
        {
            uint16_t src = scanline[sPtr];

            unsigned char r = std::min(31, (int)(data->r[iPtr]*alpha) + (src >> 11));
            unsigned char g = std::min(63, (int)(data->g[iPtr]*alpha) + ((src & 0x07E0) >> 5));
            unsigned char b = std::min(31, (int)(data->b[iPtr]*alpha) + (src & 0x001F));

            scanline[sPtr] = (r << 11) | (g << 5) | b;

            ++sPtr;
            ++iPtr;
        }
        iPtr += xwid - x_e;
        sPtr += swid - x_e;
    }
}

void drawSubtractive(int id)
{
    ImageData* data = blends[id];
    RPG::Picture* pic = data->pic;

    float alpha = 1 - 0.01*pic->transparency;

    int xwid = data->width;
    int ywid = data->height;
    int swid = RPG::screen->canvas->lineSize/2;

    int x = pic->x - xwid/2;
    int y = pic->y - ywid/2;

    int x_t;
    int y_t;

    int x_c;
    int y_c;

    int x_e;
    int y_e;

    if(x < 0)
    {
        x_t = 0;
        x_c = -x;
        int tmp = xwid + x;
        x_e = (tmp > 320)?320:tmp;
    }
    else
    {
        x_t = x;
        x_c = 0;
        x_e = (x + xwid > 320)?320-x:xwid;
    }

    if(y < 0)
    {
        y_t = 0;
        y_c = -y;
        int tmp = ywid + y;
        y_e = (tmp > 240)?240:tmp;
    }
    else
    {
        y_t = y;
        y_c = 0;
        y_e = (y + ywid > 240)?240-y:ywid;
    }

    uint16_t* scanline = RPG::screen->canvas->getScanline(0);

    int sPtr = x_t + y_t * swid;
    int iPtr = x_c + y_c * xwid;

    for(int i = 0; i < y_e; ++i)
    {
        for(int j = 0; j < x_e; ++j)
        {
            uint16_t src = scanline[sPtr];

            unsigned char r = std::max(0, (src >> 11) - (int)(data->r[iPtr]*alpha));
            unsigned char g = std::max(0, ((src & 0x07E0) >> 5) - (int)(data->g[iPtr]*alpha));
            unsigned char b = std::max(0, (src & 0x001F) - (int)(data->b[iPtr]*alpha));

            scanline[sPtr] = (r << 11) | (g << 5) | b;

            ++sPtr;
            ++iPtr;
        }
        iPtr += xwid - x_e;
        sPtr += swid - x_e;
    }
}

void drawMultiplicative(int id)
{
    ImageData* data = blends[id];
    RPG::Picture* pic = data->pic;

    float alpha = 0.01*pic->transparency;

    int xwid = data->width;
    int ywid = data->height;
    int swid = RPG::screen->canvas->lineSize/2;

    int x = pic->x - xwid/2;
    int y = pic->y - ywid/2;

    int x_t;
    int y_t;

    int x_c;
    int y_c;

    int x_e;
    int y_e;

    if(x < 0)
    {
        x_t = 0;
        x_c = -x;
        int tmp = xwid + x;
        x_e = (tmp > 320)?320:tmp;
    }
    else
    {
        x_t = x;
        x_c = 0;
        x_e = (x + xwid > 320)?320-x:xwid;
    }

    if(y < 0)
    {
        y_t = 0;
        y_c = -y;
        int tmp = ywid + y;
        y_e = (tmp > 240)?240:tmp;
    }
    else
    {
        y_t = y;
        y_c = 0;
        y_e = (y + ywid > 240)?240-y:ywid;
    }

    uint16_t* scanline = RPG::screen->canvas->getScanline(0);

    int sPtr = x_t + y_t * swid;
    int iPtr = x_c + y_c * xwid;

    for(int i = 0; i < y_e; ++i)
    {
        for(int j = 0; j < x_e; ++j)
        {
            uint16_t src = scanline[sPtr];


            unsigned char r = ((src >> 11) * (int)(data->r[iPtr]+alpha*(32-data->r[iPtr])))>>5;
            unsigned char g = (((src & 0x07E0) >> 5) * (int)(data->g[iPtr]+alpha*(64-data->g[iPtr])))>>6;
            unsigned char b = ((src & 0x001F) * (int)(data->b[iPtr]+alpha*(32-data->b[iPtr])))>>5;

            scanline[sPtr] = (r << 11) | (g << 5) | b;

            ++sPtr;
            ++iPtr;
        }
        iPtr += xwid - x_e;
        sPtr += swid - x_e;
    }
}

void drawMultiplicative2x(int id)
{
    ImageData* data = blends[id];
    RPG::Picture* pic = data->pic;

    float alpha = 0.01*pic->transparency;

    int xwid = data->width;
    int ywid = data->height;
    int swid = RPG::screen->canvas->lineSize/2;

    int x = pic->x - xwid/2;
    int y = pic->y - ywid/2;

    int x_t;
    int y_t;

    int x_c;
    int y_c;

    int x_e;
    int y_e;

    if(x < 0)
    {
        x_t = 0;
        x_c = -x;
        int tmp = xwid + x;
        x_e = (tmp > 320)?320:tmp;
    }
    else
    {
        x_t = x;
        x_c = 0;
        x_e = (x + xwid > 320)?320-x:xwid;
    }

    if(y < 0)
    {
        y_t = 0;
        y_c = -y;
        int tmp = ywid + y;
        y_e = (tmp > 240)?240:tmp;
    }
    else
    {
        y_t = y;
        y_c = 0;
        y_e = (y + ywid > 240)?240-y:ywid;
    }

    uint16_t* scanline = RPG::screen->canvas->getScanline(0);

    int sPtr = x_t + y_t * swid;
    int iPtr = x_c + y_c * xwid;

    for(int i = 0; i < y_e; ++i)
    {
        for(int j = 0; j < x_e; ++j)
        {
            uint16_t src = scanline[sPtr];

            unsigned char r = std::min(31, (((src >> 11) * (int)(data->r[iPtr]+alpha*(32-data->r[iPtr])))>>5)<<1);
            unsigned char g = std::min(63, ((((src & 0x07E0) >> 5) * (int)(data->g[iPtr]+alpha*(64-data->g[iPtr])))>>6)<<1);
            unsigned char b = std::min(31, (((src & 0x001F) * (int)(data->b[iPtr]+alpha*(32-data->b[iPtr])))>>5)<<1);

            scanline[sPtr] = (r << 11) | (g << 5) | b;

            ++sPtr;
            ++iPtr;
        }
        iPtr += xwid - x_e;
        sPtr += swid - x_e;
    }
}

bool onEventCommand(RPG::EventScriptLine *scriptLine, RPG::EventScriptData *scriptData, int eventId, int pageId, int lineId, int *nextLineId)
{
    if(scriptLine->command == RPG::EVCMD_ERASE_PICTURE)
    {
        Blend::iterator itr = blends.find(scriptLine->parameter(0));
        if(itr != blends.end())
        {
            delete itr->second;
            blends.erase(itr);
        }
        Active::iterator tmp = active.find(scriptLine->parameter(0));
        active.erase(tmp);
    }
    if(scriptLine->command == RPG::EVCMD_SHOW_PICTURE)
    {
        RPG::variables[7] = scriptLine->parameter(0);
        Blend::iterator itr = blends.find(scriptLine->parameter(0));
        if(itr != blends.end())
        {
            delete itr->second;
            blends.erase(itr);
        }
        Active::iterator tmp = active.find(scriptLine->parameter(0));
        if(tmp != active.end())
            active.erase(tmp);
    }
    return true;
}

bool onDrawPicture(RPG::Picture* picture)
{
    int id = picture->id;
    Active::iterator act = active.find(id);
    if(act == active.end())
    {
        std::string type = picture->filename;
        if(type.size() < 4)
        {
            active[id] = 0;
            return true;
        }
        type = type.substr(type.size()-4,4);

        if(!type.compare(".add"))
        {
            active[id] = 1;
            blends[id] = new ImageData(picture);
            drawAdditive(id);
            return true;
        }
        if(!type.compare(".sub"))
        {
            active[id] = 2;
            blends[id] = new ImageData(picture);
            drawSubtractive(id);
            return true;
        }
        if(!type.compare(".mul"))
        {
            active[id] = 3;
            blends[id] = new ImageData(picture);
            drawMultiplicative(id);
            return true;
        }
        if(!type.compare(".ml2"))
        {
            active[id] = 4;
            blends[id] = new ImageData(picture);
            drawMultiplicative2x(id);
            return true;
        }
        active[id] = 0;
        return true;
    }
    else
    {
        Blend::iterator itr = blends.find(picture->id);
        if(itr != blends.end())
        {
            switch(act->second)
            {
                case 1:
                    drawAdditive(id);
                    return true;
                case 2:
                    drawSubtractive(id);
                    return true;
                case 3:
                    drawMultiplicative(id);
                    return true;
                case 4:
                    drawMultiplicative2x(id);
                    return true;
            }
        }
    }
    return true;
}
